import datetime as dt

import numpy as np


def calculate_crop_evapotranspiration(**kwargs):
    model = CropEvapotranspiration(**kwargs)
    model.calculate()


class CropEvapotranspiration(object):
    def __init__(self, **kwargs):
        for arg in kwargs:
            setattr(self, arg, kwargs[arg])

    def calculate(self):
        self.timeseries["kc"] = 0
        self._calculate_kc_for_unplanted_stage()
        self._calculate_kc_for_init_stage()
        self._calculate_kc_for_dev_stage()
        self._calculate_kc_for_mid_stage()
        self._calculate_kc_for_late_stage()
        self.timeseries["crop_evapotranspiration"] = (
            self.timeseries["ref_evapotranspiration"] * self.timeseries["kc"]
        )

    def _calculate_kc_for_unplanted_stage(self):
        end = self.planting_date - dt.timedelta(days=1)
        self.timeseries.loc[:end, "kc"] = self.kc_unplanted

    def _calculate_kc_for_init_stage(self):
        start = self.planting_date
        end = self.planting_date + dt.timedelta(days=self.init - 1)
        self.timeseries.loc[start:end, "kc"] = self.kc_ini

    def _calculate_kc_for_dev_stage(self):
        start = self.planting_date + dt.timedelta(days=self.init)
        end = start + dt.timedelta(days=self.dev - 1)
        dev_kcs = np.linspace(self.kc_ini, self.kc_mid, num=self.dev + 1)[1:]
        self.timeseries.loc[start:end, "kc"] = dev_kcs

    def _calculate_kc_for_mid_stage(self):
        start = self.planting_date + dt.timedelta(days=self.init + self.dev)
        end = start + dt.timedelta(days=self.mid - 1)
        self.timeseries.loc[start:end, "kc"] = self.kc_mid

    def _calculate_kc_for_late_stage(self):
        start = self.planting_date + dt.timedelta(days=self.init + self.dev + self.mid)
        end = start + dt.timedelta(days=self.late - 1)
        late_kcs = np.linspace(self.kc_mid, self.kc_end, num=self.late + 1)[1:]
        self.timeseries.loc[start:end, "kc"] = late_kcs



import datetime as dt
from unittest import TestCase

import numpy as np
import pandas as pd

from swb import calculate_crop_evapotranspiration


class CalculateCropEvapotranspirationTestCase(TestCase):
    """Functional testing of calculate_crop_evapotranspiration().

    We use the example of FAO56 (Box 15, Figure 36, Example 28, pp. 130-133) for
    testing. However we consider a planting date of one day earlier than what the
    example says, otherwise results don't agree. We also use 10 additional days before
    planting in order to test kc_unplanted.

    Start of time series: 1974-05-12
    Planting date: 1974-05-22
    Total crop lifetime: 100 days (1974-05-22 to 1974-08-29)
    """

    def setUp(self):
        self.timeseries = pd.DataFrame(
            data={"ref_evapotranspiration": np.full(110, 3.14)},
            index=pd.date_range("1974-05-12", periods=110),
        )
        calculate_crop_evapotranspiration(
            timeseries=self.timeseries,
            planting_date=dt.date(1974, 5, 22),
            kc_unplanted=0.1,
            kc_ini=0.15,
            kc_mid=1.19,
            kc_end=0.35,
            init=25,
            dev=25,
            mid=30,
            late=20,
        )

    def test_kc_start(self):
        self.assertAlmostEqual(self.timeseries.loc["1974-05-12"]["kc"], 0.1, places=2)

    def test_kc_end_before_planting(self):
        self.assertAlmostEqual(self.timeseries.loc["1974-05-21"]["kc"], 0.1, places=2)

    def test_kc_start_of_init(self):
        self.assertAlmostEqual(self.timeseries.loc["1974-05-22"]["kc"], 0.15, places=2)

    def test_kc_end_of_init(self):
        self.assertAlmostEqual(self.timeseries.loc["1974-06-15"]["kc"], 0.15, places=2)

    def test_kc_june_20(self):
        # This result is indicated in a paragraph just below Figure 36
        self.assertAlmostEqual(self.timeseries.loc["1974-06-20"]["kc"], 0.36, places=2)

    def test_kc_on_day_40(self):
        # Example 28 p. 133
        self.assertAlmostEqual(self.timeseries.iloc[9 + 40]["kc"], 0.77, places=2)

    def test_kc_on_day_70(self):
        # Example 28 p. 133
        self.assertAlmostEqual(self.timeseries.iloc[9 + 70]["kc"], 1.19, places=2)

    def test_kc_on_day_95(self):
        # Example 28 p. 133
        self.assertAlmostEqual(self.timeseries.iloc[9 + 95]["kc"], 0.56, places=2)

    def test_kc_on_last_day(self):
        self.assertAlmostEqual(self.timeseries.iloc[9 + 100]["kc"], 0.35, places=2)

    def test_crop_evapotranspiration(self):
        # The hard part is to calculate Kc. The crop evapotranspiration is trivial.
        # We've specified a constant reference evapotranspiration of 3.14, and all we
        # need to do is check one day and verify that the multiplication with Kc has
        # been done correctly. We use day 40.
        self.assertAlmostEqual(
            self.timeseries.iloc[9 + 40]["crop_evapotranspiration"],
            3.14 * 0.774,
            places=2,
        )



from unittest import TestCase

import pandas as pd

from swb import SoilWaterBalance, calculate_soil_water


class SimpleMethodsTestCase(TestCase):
    def setUp(self):
        self.swb = SoilWaterBalance(
            theta_s=0.425,
            theta_fc=0.287,
            theta_wp=0.14,
            zr=0.5,
            zr_factor=1000,
            p=0.5,
            draintime=16.2,
            timeseries=None,
            theta_init=None,
            mif=None,
        )

    def test_taw(self):
        self.assertAlmostEqual(self.swb.taw, 73.5)

    def test_raw(self):
        self.assertAlmostEqual(self.swb.raw, 36.75)

    def test_dr_from_theta(self):
        self.assertAlmostEqual(self.swb.dr_from_theta(0.277326), 4.837, places=3)

    def test_theta_from_dr(self):
        self.assertAlmostEqual(self.swb.theta_from_dr(4.837), 0.277, places=3)

    def test_ks(self):
        self.assertAlmostEqual(self.swb.ks(37.746), 0.973, places=3)

    def test_ks_when_above_threshold(self):
        self.assertAlmostEqual(self.swb.ks(30), 1)

    def test_dp(self):
        self.assertAlmostEqual(self.swb.dp(0.311, 0), 0.741, places=3)

    def test_dp_zero(self):
        self.assertEqual(self.swb.dp(0.186, 0), 0)

    def test_ro_at_saturation(self):
        # When we are already saturated (0.425), all precipitation (15 mm) runs off
        self.assertAlmostEqual(self.swb.ro(15, 0.425), 15)

    def test_ro_zero(self):
        # When the soil is practically dry and 2 mm falls, nothing runs off
        self.assertEqual(self.swb.ro(2, 0.14), 0)


class ModelRunWithActualNetIrrigationTestCase(TestCase):
    def setUp(self):
        data = {
            "effective_precipitation": [0, 4.8],
            "actual_net_irrigation": [0, 0],
            "crop_evapotranspiration": [3.871, 3.990],
        }
        self.df = pd.DataFrame(data, index=pd.date_range("2019-03-07", periods=2))
        calculate_soil_water(
            theta_s=0.425,
            theta_fc=0.287,
            theta_wp=0.14,
            zr=0.5,
            zr_factor=1000,
            p=0.5,
            draintime=16.2,
            timeseries=self.df,
            theta_init=0.252784,
            mif=None,
        )

    def test_dr1(self):
        self.assertAlmostEqual(self.df.iloc[0]["dr"], 20.979, places=3)

    def test_dr2(self):
        self.assertAlmostEqual(self.df.iloc[1]["dr"], 20.169, places=3)

    def test_theta1(self):
        self.assertAlmostEqual(self.df.iloc[0]["theta"], 0.245, places=3)

    def test_theta2(self):
        self.assertAlmostEqual(self.df.iloc[1]["theta"], 0.247, places=3)

    def test_ks1(self):
        self.assertAlmostEqual(self.df.iloc[0]["ks"], 1)

    def test_ks2(self):
        self.assertAlmostEqual(self.df.iloc[1]["ks"], 1)

    def test_recommended_net_irrigation1(self):
        self.assertAlmostEqual(self.df.iloc[0]["recommended_net_irrigation"], 0)

    def test_recommended_net_irrigation2(self):
        self.assertAlmostEqual(self.df.iloc[1]["recommended_net_irrigation"], 0)


class ModelRunWithNonzeroKsTestCase(TestCase):
    def setUp(self):
        data = {
            "effective_precipitation": [0, 0],
            "actual_net_irrigation": [0, 0],
            "crop_evapotranspiration": [3.724, 3.906],
        }
        self.df = pd.DataFrame(data, index=pd.date_range("2019-08-18", periods=2))
        calculate_soil_water(
            theta_s=0.425,
            theta_fc=0.287,
            theta_wp=0.14,
            zr=0.5,
            zr_factor=1000,
            p=0.5,
            draintime=16.2,
            timeseries=self.df,
            theta_init=0.218956,
            mif=1,
        )

    def test_dr1(self):
        self.assertAlmostEqual(self.df.iloc[0]["dr"], 37.746, places=3)

    def test_dr2(self):
        self.assertAlmostEqual(self.df.iloc[1]["dr"], 41.546, places=3)

    def test_theta1(self):
        self.assertAlmostEqual(self.df.iloc[0]["theta"], 0.212, places=3)

    def test_theta2(self):
        self.assertAlmostEqual(self.df.iloc[1]["theta"], 0.204, places=3)

    def test_ks1(self):
        self.assertAlmostEqual(self.df.iloc[0]["ks"], 1)

    def test_ks2(self):
        self.assertAlmostEqual(self.df.iloc[1]["ks"], 0.973, places=3)


class ModelRunWithAutoApplyIrrigationTestCase(TestCase):
    def setUp(self):
        data = {
            "effective_precipitation": [0, 0, 0],
            "actual_net_irrigation": [True, True, True],
            "crop_evapotranspiration": [3.647, 3.822, 3.885],
        }
        self.df = pd.DataFrame(data, index=pd.date_range("2019-03-07", periods=3))
        calculate_soil_water(
            theta_s=0.425,
            theta_fc=0.287,
            theta_wp=0.14,
            zr=0.5,
            zr_factor=1000,
            p=0.5,
            draintime=16.2,
            timeseries=self.df,
            theta_init=0.223646,
            mif=1.0,
        )

    def test_dr1(self):
        self.assertAlmostEqual(self.df.iloc[0]["dr"], 35.324, places=3)

    def test_dr2(self):
        self.assertAlmostEqual(self.df.iloc[1]["dr"], 0, places=3)

    def test_dr3(self):
        self.assertAlmostEqual(self.df.iloc[2]["dr"], 3.885, places=3)

    def test_theta1(self):
        self.assertAlmostEqual(self.df.iloc[0]["theta"], 0.216, places=3)

    def test_theta2(self):
        self.assertAlmostEqual(self.df.iloc[1]["theta"], 0.287, places=3)

    def test_theta3(self):
        self.assertAlmostEqual(self.df.iloc[2]["theta"], 0.279, places=3)

    def test_recommended_net_irrigation1(self):
        self.assertAlmostEqual(self.df.iloc[0]["recommended_net_irrigation"], 0)

    def test_recommended_net_irrigation2(self):
        self.assertAlmostEqual(
            self.df.iloc[1]["recommended_net_irrigation"], 39.146, places=3
        )

    def test_recommended_net_irrigation3(self):
        self.assertAlmostEqual(self.df.iloc[2]["recommended_net_irrigation"], 0)


class MifTestCase(TestCase):
    def setUp(self):
        data = {
            "effective_precipitation": [0, 0, 0],
            "actual_net_irrigation": [True, True, True],
            "crop_evapotranspiration": [3.647, 3.822, 3.885],
        }
        self.df = pd.DataFrame(data, index=pd.date_range("2019-03-07", periods=3))
        calculate_soil_water(
            theta_s=0.425,
            theta_fc=0.287,
            theta_wp=0.14,
            zr=0.5,
            zr_factor=1000,
            p=0.5,
            draintime=16.2,
            timeseries=self.df,
            theta_init=0.223646,
            mif=0.5,
        )

    def test_dr1(self):
        self.assertAlmostEqual(self.df.iloc[0]["dr"], 35.324, places=3)

    def test_dr2(self):
        self.assertAlmostEqual(self.df.iloc[1]["dr"], 19.573, places=3)

    def test_dr3(self):
        self.assertAlmostEqual(self.df.iloc[2]["dr"], 23.458, places=3)

    def test_recommended_net_irrigation1(self):
        self.assertAlmostEqual(self.df.iloc[0]["recommended_net_irrigation"], 0)

    def test_recommended_net_irrigation2(self):
        self.assertAlmostEqual(
            self.df.iloc[1]["recommended_net_irrigation"], 19.573, places=3
        )

    def test_recommended_net_irrigation3(self):
        self.assertAlmostEqual(self.df.iloc[2]["recommended_net_irrigation"], 0)


class ModelRunWithDpTestCase(TestCase):
    def setUp(self):
        data = {
            "effective_precipitation": [0, 5.544, 2.664],
            "actual_net_irrigation": [0, 0, 0],
            "crop_evapotranspiration": [1.043, 0.952, 0.868],
        }
        self.df = pd.DataFrame(data, index=pd.date_range("2016-03-10", periods=3))
        calculate_soil_water(
            theta_s=0.425,
            theta_fc=0.287,
            theta_wp=0.14,
            zr=0.5,
            zr_factor=1000,
            p=0.5,
            draintime=16.3,
            timeseries=self.df,
            theta_init=0.284732,
            mif=1.0,
        )

    def test_dr1(self):
        self.assertAlmostEqual(self.df.iloc[0]["dr"], 2.177, places=3)

    def test_dr2(self):
        self.assertAlmostEqual(self.df.iloc[1]["dr"], -2.208, places=3)

    def test_dr3(self):
        self.assertAlmostEqual(self.df.iloc[2]["dr"], -3.706, places=3)



import numpy as np


def calculate_soil_water(**kwargs):
    model = SoilWaterBalance(**kwargs)
    model.calculate_timeseries()
    return {"raw": model.raw, "taw": model.taw, "timeseries": kwargs["timeseries"]}


class SoilWaterBalance(object):
    # Symbols we use:
    # theta_s - Water content at saturation
    # theta_fc - Water content at field capacity
    # theta_wp - Water content at wilting point
    # theta_init - Initial water content
    # zr - Root zone depth
    # zr_factor - Unit change factor for reducing zr to the same unit as the time series
    # kc - Crop coefficient
    # p - Soil water depletion fraction for no stress
    # draintime - Days the soil needs to go from theta_s to theta_fc
    # raw - Readily available water
    # taw - Total available water
    # effective_precipitation - Effective precipitation
    # evapotranspiration - Reference evapotranspiration
    # crop_evapotranspiration - The reference evaporation multiplied by kc
    # actual_net_irrigation - The actual supplied net irrigation
    # recommended_net_irrigation - The calculated required net irrigation
    # ro - Surface runoff
    # cr - Capillary rise
    # dp - Deep percolation
    # dr - Root zone depletion
    # mif - Malamos irrigation factor

    def __init__(self, **kwargs):
        self.theta_s = kwargs["theta_s"]
        self.theta_fc = kwargs["theta_fc"]
        self.theta_wp = kwargs["theta_wp"]
        self.zr = kwargs["zr"]
        self.zr_factor = kwargs["zr_factor"]
        self.p = kwargs["p"]
        self.draintime = kwargs["draintime"]
        self.timeseries = kwargs["timeseries"]
        self.theta_init = kwargs["theta_init"]
        self.mif = kwargs["mif"]

        self.taw = (self.theta_fc - self.theta_wp) * self.zr * self.zr_factor
        self.raw = self.p * self.taw

    def calculate_timeseries(self):
        # Add columns to self.timeseries
        self.timeseries["dr"] = np.nan
        self.timeseries["theta"] = np.nan
        self.timeseries["ks"] = np.nan
        self.timeseries["recommended_net_irrigation"] = np.nan

        # Loop and perform the calculation
        theta_prev = self.theta_init
        dr_prev = self.dr_from_theta(theta_prev)
        for date in self.timeseries.index:
            row = self.timeseries.loc[date]
            ks = self.ks(dr_prev)
            dr_without_irrig = self.dr_without_irrig(dr_prev, theta_prev, ks, row)

            ani = row["actual_net_irrigation"]
            auto_apply_irrigation = type(ani) == np.bool_ and ani
            assumed_net_irrigation = 0 if auto_apply_irrigation else ani
            dr = dr_without_irrig - assumed_net_irrigation
            recommended_net_irrigation = dr * self.mif if dr > self.raw else 0
            if auto_apply_irrigation and dr_without_irrig > self.raw:
                dr = dr_without_irrig - recommended_net_irrigation

            theta = self.theta_from_dr(dr)
            self.timeseries.at[date, "dr"] = dr
            self.timeseries.at[date, "theta"] = theta
            self.timeseries.at[date, "ks"] = ks
            self.timeseries.at[
                date, "recommended_net_irrigation"
            ] = recommended_net_irrigation
            theta_prev = theta
            dr_prev = dr

    def dr_from_theta(self, theta):
        return (self.theta_fc - theta) * self.zr * self.zr_factor

    def theta_from_dr(self, dr):
        return self.theta_fc - dr / (self.zr * self.zr_factor)

    def ks(self, dr):
        result = (self.taw - dr) / ((1 - self.p) * self.taw)
        return min(result, 1)

    def ro(self, effective_precipitation, theta_prev):
        result = (
            effective_precipitation
            + (theta_prev - self.theta_s) * self.zr * self.zr_factor
        )
        return max(result, 0)

    def dp(self, theta_prev, peff):
        theta_prev_mm = theta_prev * self.zr * self.zr_factor
        theta_fc_mm = self.theta_fc * self.zr * self.zr_factor
        excess_water = theta_prev_mm - theta_fc_mm + peff
        return max(excess_water, 0) / self.draintime

    def dr_without_irrig(self, dr_prev, theta_prev, ks, row):
        # "row" is a single row from self.timeseries
        return (
            dr_prev
            - (
                row["effective_precipitation"]
                - self.ro(row["effective_precipitation"], theta_prev)
            )
            + row["crop_evapotranspiration"] * ks
            + self.dp(theta_prev, row["effective_precipitation"])
        )



#!/usr/bin/env python

from setuptools import find_packages, setup

with open("README.rst") as readme_file:
    readme = readme_file.read()

with open("HISTORY.rst") as history_file:
    history = history_file.read()

requirements = ["pandas>=0.20"]

setup_requirements = []

test_requirements = []

setup(
    name="swb",
    author="Antonis Christofides",
    author_email="antonis@antonischristofides.com",
    classifiers=[
        "Development Status :: 2 - Pre-Alpha",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
        "Natural Language :: English",
        "Programming Language :: Python :: 3.5",
        "Programming Language :: Python :: 3.6",
    ],
    description="Calculation of soil water balance",
    install_requires=requirements,
    license="GNU General Public License v3",
    long_description=readme + "\n\n" + history,
    include_package_data=True,
    packages=find_packages(include=["swb"]),
    setup_requires=setup_requirements,
    test_suite="tests",
    tests_require=test_requirements,
    url="https://github.com/openmeteo/swb",
    version="0.3.1",
    zip_safe=False,
)
